import numpy as np
import vigra

from extensions.chaeo.accessors import MonoPatchStack
from extensions.ilastik.models import IlastikObjectClassifierFromSegmentationModel
from model_server.accessors import InMemoryDataAccessor


class PatchStackObjectClassifier(IlastikObjectClassifierFromSegmentationModel):
    """
    Wrap ilastik object classification for inputs comprising raw image and binary segmentation masks, both represented
    as time-series images where each frame contains only one object.
    """


    def infer(self, input_acc: MonoPatchStack, segmentation_acc: MonoPatchStack) -> (np.ndarray, dict):
        assert segmentation_acc.is_mask()
        assert input_acc.chroma == 1

        tagged_input_data = vigra.taggedView(input_acc.make_tczyx(), 'tczyx')
        tagged_seg_data = vigra.taggedView(segmentation_acc.make_tczyx(), 'tczyx')

        dsi = [
            {
                'Raw Data': self.PreloadedArrayDatasetInfo(preloaded_array=tagged_input_data),
                'Segmentation Image': self.PreloadedArrayDatasetInfo(preloaded_array=tagged_seg_data),
            }
        ]

        obmaps = self.shell.workflow.batchProcessingApplet.run_export(dsi, export_to_array=True)  # [z x h x w x n]

        assert len(obmaps) == 1, 'ilastik generated more than one object map'

        # for some reason ilastik scrambles these axes to Z(1)YX(1)
        assert obmaps[0].shape == (input_acc.nz, 1, input_acc.hw[0], input_acc.hw[1], 1)
        yxz = np.moveaxis(
            obmaps[0][:, 0, :, :, 0],
            [1, 2, 0],
            [0, 1, 2]
        )

        assert yxz.shape[0:2] == input_acc.hw
        assert yxz.shape[2] == input_acc.nz
        return MonoPatchStack(data=yxz), {'success': True}